from pathlib import Path
from typing import Optional

from yes3.s3 import is_s3_url, S3Location
from yes3.caching import CacheCore, LocalDiskCache, MemoryCache, MultiCache, S3Cache


def setup_single_cache(
        path: Optional[str | Path | S3Location | CacheCore] = None,
        in_memory=False,
        rebuild_missing_metadata=False,
) -> CacheCore:
    if isinstance(path, CacheCore):
        cache = path
    elif isinstance(path, S3Location) or (isinstance(path, str) and is_s3_url(path)):
        cache = S3Cache.create(path, rebuild_missing_meta=rebuild_missing_metadata)
    elif isinstance(path, Path) or isinstance(path, str):
        cache = LocalDiskCache.create(path, rebuild_missing_meta=rebuild_missing_metadata)
    elif not in_memory:
        raise TypeError('`path` must be a Cache, local path, or s3 location')
    else:
        cache = MemoryCache()
    return cache


def setup_cache(*paths, in_memory=False, sync=False, rebuild_missing_metadata=False) -> CacheCore | None:
    caches = []
    if in_memory:
        caches.append(setup_single_cache(in_memory=in_memory))
    for path in paths:
        if path is not None:
            if not (isinstance(path, str) or isinstance(path, CacheCore)) and hasattr(path, '__len__'):
                for p in path:
                    caches.append(setup_single_cache(p, rebuild_missing_metadata=rebuild_missing_metadata))
            else:
                caches.append(setup_single_cache(path, rebuild_missing_metadata=rebuild_missing_metadata))
    if len(caches) == 0:
        return None
    elif len(caches) == 1:
        return caches[0]
    else:
        return MultiCache(caches, sync_all=sync)



import os
from functools import partial
from pathlib import Path
from typing import Optional, Self

from yes3 import s3
from yes3.caching.base import Cache, CacheDictCatalog, CachedItemMeta, CacheReaderWriter
from yes3.s3 import S3Location


def _with_ext(path: S3Location, ext: Optional[str]) -> S3Location:
    if ext is None:
        return path
    if not ext.startswith('.'):
        ext = f'.{ext}'
    key = path.key
    if key.endswith(ext):
        return path
    else:
        return type(path)(path.bucket, key + ext, path.region)


class S3ReaderWriter(CacheReaderWriter):
    def __init__(
            self,
            path: str | S3Location,
            file_type: str = 'pkl',
            meta_file_type: str = 'json',
            meta_ext: str = 'meta',
            progress=None,
    ):
        self.path = S3Location(path)
        self._file_type = file_type
        self._meta_file_type = meta_file_type
        self._meta_ext = meta_ext
        self._progress = progress

    def key2path(self, key: str, meta=False) -> S3Location:
        if meta:
            return _with_ext(self.path.join(key), self._meta_ext)
        else:
            return _with_ext(self.path.join(key), self._file_type)

    def path2key(self, path: str | S3Location) -> str:
        path = S3Location(path)
        filename = path.s3_uri.split(self.path.s3_uri, maxsplit=1)[-1].lstrip('/')
        key, ext = os.path.splitext(filename)
        return key

    def read(self, key: str, file_type=None):
        file_type = file_type or self._file_type
        path = self.key2path(key)
        print(f"Reading cached item '{key}' at {path.s3_uri}")
        return s3.read(path, file_type=file_type)

    def _build_meta(self, path, key=None) -> CachedItemMeta:
        if key is None:
            key = self.path2key(path)
        rel_path = Path(path.key).relative_to(Path(self.path.key))
        obj_meta = path.get_object_metadata()
        return CachedItemMeta(
            key=key,
            path=str(rel_path),
            size=obj_meta.size,
            timestamp=obj_meta.last_modified,
        )

    def get_meta(self, key: str, rebuild=False, file_type=None) -> CachedItemMeta:
        if rebuild:
            obj_path = self.key2path(key)
            meta_path = self.key2path(key, meta=True)
            meta = self._build_meta(path=obj_path, key=key)
            s3.write_to_s3(meta.to_dict(), meta_path, file_type=self._meta_file_type)
        else:
            meta_path = self.key2path(key, meta=True)
            meta_dict = s3.read(meta_path, file_type=self._meta_file_type)
            meta = CachedItemMeta(**meta_dict)
        return meta

    def write(self, key: str, obj, meta: Optional[CachedItemMeta] = None, file_type=None) -> CachedItemMeta:
        file_type = file_type or self._file_type
        obj_path = self.key2path(key)
        print(f"Caching item '{key}' at {obj_path.s3_uri}")
        s3.write_to_s3(obj, obj_path, file_type=file_type, progress=self._progress)

        meta_path = self.key2path(key, meta=True)
        if meta is None:
            meta = self._build_meta(path=obj_path, key=key)
        s3.write_to_s3(meta.to_dict(), meta_path, file_type=self._meta_file_type)
        return meta

    def delete(self, key: str, meta_only=False):
        path = self.key2path(key)
        meta_path = self.key2path(key, meta=True)
        if meta_only:
            print(f"Deleting cached item '{key}' metadata at {meta_path.s3_uri}")
        else:
            print(f"Deleting cached item '{key}' at {path.s3_uri}")
            s3.delete(path)
        s3.delete(meta_path)


class S3Cache(Cache):
    @staticmethod
    def _build_catalog_dict(reader_writer: S3ReaderWriter, rebuild_missing_meta=False) -> dict:
        catalog_dict = {}
        locations = s3.list_objects(reader_writer.path)
        meta_locs = [loc for loc in locations if loc.key.endswith(reader_writer._meta_ext)]
        data_locs = [loc for loc in locations if loc not in meta_locs]
        data_map = {reader_writer.path2key(loc): loc for loc in data_locs}
        meta_map = {reader_writer.path2key(loc): loc for loc in meta_locs}
        if data_map.keys() != meta_map.keys():
            if rebuild_missing_meta:
                print(f'WARNING: data and metadata files are not aligned for cache at {reader_writer.path}, '
                      'rebuilding missing metadata files')
            else:
                raise RuntimeError(f'data and metadata files are not aligned for cache at {reader_writer.path}')
        for key in data_map.keys():
            catalog_dict[key] = reader_writer.get_meta(key, rebuild=(key not in meta_map and rebuild_missing_meta))
        if len(catalog_dict.keys()) > 0:
            print(f'{len(catalog_dict.keys())} cached items discovered at {reader_writer.path.s3_uri}')
        return catalog_dict

    @classmethod
    def create(
            cls,
            path: str | S3Location,
            file_type=None,
            meta_file_type=None,
            meta_ext=None,
            reader_writer: Optional[CacheReaderWriter] = None,
            rebuild_missing_meta=False,
            **kwargs,
    ):
        if reader_writer is None:
            rw_kwargs = {}
            if file_type is not None:
                rw_kwargs['file_type'] = file_type
            if meta_file_type is not None:
                rw_kwargs['meta_file_type'] = meta_file_type
            if meta_ext is not None:
                rw_kwargs['meta_ext'] = meta_ext
            reader_writer = S3ReaderWriter(path, **rw_kwargs)
        catalog_builder = partial(cls._build_catalog_dict, reader_writer=reader_writer,
                                  rebuild_missing_meta=rebuild_missing_meta)
        catalog = CacheDictCatalog(catalog_builder=catalog_builder)
        return cls(catalog, reader_writer, **kwargs)

    @property
    def path(self) -> S3Location:
        return self._reader_writer.path

    def subcache(self, rel_path: str) -> Self:
        path = self.path / rel_path
        kwargs = dict(active=self.is_active(), read_only=self.is_read_only())
        return self.create(path, reader_writer=self._reader_writer, **kwargs)

    def clear(self, force=False) -> Self:
        if self.is_active() and len(self.keys()) > 0:
            if not force:
                raise RuntimeError(f'Clearing this {type(self).__name__} ({self.path.s3_uri}) requires specifying '
                                   'force=True')
            print(f'Deleting {len(self.keys())} item(s) from cache at {self.path.s3_uri}')
            for key in self.keys():
                self.remove(key)
            new_cache = type(self).create(self.path, reader_writer=self._reader_writer)
            self.__init__(new_cache._catalog, new_cache._reader_writer, active=self._active, read_only=self._read_only)
        return self

    def clear_meta(self, force=False) -> Self:
        if self.is_active() and len(self.keys()) > 0:
            if not force:
                raise RuntimeError(f'Clearing this {type(self).__name__} metadata ({self.path.s3_uri}) requires '
                                   'specifying force=True')
            print(f'Deleting {len(self.keys())} item(s) from cache at {self.path.s3_uri}')
            for key in self.keys():
                self.remove(key, meta_only=True)
        return self

    def _repr_params(self) -> list[str]:
        params = super()._repr_params()
        params.insert(0, self.path.s3_uri)
        return params



import os
import shutil
import unittest
from pathlib import Path
from typing import List

from moto import mock_aws

from yes3 import s3, S3Location
from yes3.s3 import S3Object
from yes3.utils.testing import get_arg_parser, run_tests

TEST_LOCAL_DIR = Path('_tmp_test_dir_')
LOCAL_COPY_DIR = str(TEST_LOCAL_DIR) + 'copy_/'
TEST_BUCKET = 'mock-bucket'
TEST_S3_DIR = f's3://{TEST_BUCKET}/unit-tests/'
VERBOSE = False

TEST_FILE_CONTENTS = {
    TEST_LOCAL_DIR / 'empty_file': None,
    TEST_LOCAL_DIR / 'file1.txt': 'file1 contents',
    TEST_LOCAL_DIR / 'dir1/file1.1.txt': 'file1.1 contents',
    TEST_LOCAL_DIR / 'dir1/file1.2': None,
    TEST_LOCAL_DIR / 'dir1/dir1.2/file2.1': None,
    TEST_LOCAL_DIR / 'dir1/dir1.2/file2.2': 'file2.2 contents',
}


def _vprint(*args, **kwargs):
    if VERBOSE:
        print(*args, **kwargs)


def _cleanup_local():
    def rm(path):
        path = Path(path)
        if path.exists():
            _vprint(f'Deleting {path}')
            shutil.rmtree(path)
        assert path.exists() is False

    _vprint('----- Cleaning up local files -----')
    rm(TEST_LOCAL_DIR)
    rm(LOCAL_COPY_DIR)
    _vprint()


def _create_test_files() -> List[os.PathLike]:
    paths = []
    for path, content in TEST_FILE_CONTENTS.items():
        os.makedirs(path.parent, exist_ok=True)
        if content is not None:
            with open(path, 'w') as f:
                f.write(content)
        else:
            path.touch(exist_ok=True)
        paths.append(path)
    return paths


def _test_single_uploads():
    s3_dir_loc = s3.S3Location(TEST_S3_DIR)

    if s3_dir_loc.exists():
        s3.delete(s3_dir_loc, recursive=True)
    assert s3_dir_loc.exists() is False
    assert s3_dir_loc.is_dir() is False
    assert s3_dir_loc.is_dir_path() is True

    # Test single file uploads
    file_loc = s3.upload(TEST_LOCAL_DIR / 'empty_file', s3_dir_loc)
    assert file_loc.key == s3_dir_loc.join('empty_file').key
    assert file_loc.exists() is True
    assert file_loc.is_object() is True
    assert file_loc.is_dir() is False

    file_loc = s3.upload(TEST_LOCAL_DIR / 'dir1/file1.1.txt', s3_dir_loc.join('dir1/file'))
    dir_loc = s3_dir_loc.join('dir1')
    assert file_loc.key == dir_loc.join('file').key
    assert file_loc.exists() is True
    assert file_loc.is_object() is True
    assert file_loc.is_dir() is False
    assert dir_loc.exists() is True
    assert dir_loc.is_dir() is True
    assert dir_loc.is_object() is False


def _test_delete():
    s3_dir_loc = s3.S3Location(TEST_S3_DIR)

    file_loc = s3_dir_loc.join('empty_file')
    s3.delete(file_loc)
    assert file_loc.exists() is False

    s3.delete(s3_dir_loc.bucket, s3_dir_loc.join('dir1', 'file').key)
    assert len(s3.list_objects(TEST_S3_DIR)) == 0


def _test_recursive_uploads():
    s3_dir_loc = s3.S3Location(TEST_S3_DIR)
    file_locs = s3.upload(TEST_LOCAL_DIR, s3_dir_loc, recursive=True)
    assert len(file_locs) == len(TEST_FILE_CONTENTS)


def _test_list_objects():
    dir12 = s3.S3Location(TEST_S3_DIR).join('dir1/dir1.2')
    dir12_locs = s3.list_objects(dir12, return_metadata=False)
    for obj in dir12_locs:
        _vprint(f'Found S3 object: {obj}')
    assert len(dir12_locs) == 2
    assert set(loc.split_key()[1] for loc in dir12_locs) == {'file2.1', 'file2.2'}

    dir12_objs = s3.list_objects(dir12, return_metadata=True)
    assert len(dir12_objs) == 2
    assert all(isinstance(obj, S3Object) for obj in dir12_objs)
    assert isinstance(dir12_objs[0].location.get_object_metadata(), S3Object)


def _test_list_dir():
    dir_objs = s3.list_dir(TEST_S3_DIR, return_metadata=True, depth=1)
    assert len(dir_objs) == 3
    s3_objs = [o for o in dir_objs if isinstance(o, s3.S3Object)]
    for obj in s3_objs:
        _vprint(f'Found S3 object: {obj}')
    s3_pfxs = [o for o in dir_objs if isinstance(o, s3.S3Prefix)]
    for pfx in s3_pfxs:
        pfx.count_objects()
        _vprint(f'Found S3 prefix: {pfx}')
    assert len(s3_objs) == 2
    assert all(obj.location.is_object() for obj in s3_objs)
    assert len(s3_pfxs) == 1
    assert all(pfx.location.is_dir() for pfx in s3_pfxs)

    # List two levels down; both dir and its contents listed
    dir_objs = s3.list_dir(TEST_S3_DIR, depth=2)
    assert len(dir_objs) == 6
    _vprint()
    for loc in dir_objs:
        _vprint(f'Found S3 object: {loc}')

    # List all levels down
    dir_objs = s3.list_dir(TEST_S3_DIR, recursive=True)
    assert len(dir_objs) == 8
    _vprint()
    for loc in dir_objs:
        _vprint(f'Found S3 object: {loc}')


def _test_download():
    s3_dir_loc = s3.S3Location(TEST_S3_DIR)

    dwnld_path = s3.download(s3_dir_loc.join('empty_file'), LOCAL_COPY_DIR)
    assert dwnld_path == str(Path(LOCAL_COPY_DIR).resolve() / 'empty_file')
    assert os.path.exists(dwnld_path) is True
    os.remove(dwnld_path)

    dwnld_paths = s3.download(s3_dir_loc, LOCAL_COPY_DIR, recursive=True)
    assert len(dwnld_paths) == len(TEST_FILE_CONTENTS)

    _vprint(f'Deleting {LOCAL_COPY_DIR}')
    shutil.rmtree(LOCAL_COPY_DIR)
    assert os.path.exists(LOCAL_COPY_DIR) is False


def _test_write_read():
    contents = {'a': 42, 'b': list(range(10)), 'c': None}
    s3_path = s3.S3Location(TEST_S3_DIR).join('test_write_file.pkl')
    s3.write_to_s3(contents, s3_path)
    read_contents = s3.read(s3_path)
    assert contents == read_contents


class TestS3Utils(unittest.TestCase):

    @mock_aws
    def test_s3_location(self):
        # moto (aws mock) requires the bucket be created before use
        s3._client.create_bucket(Bucket=TEST_BUCKET)

        # Test with bucket only
        loc = S3Location(TEST_BUCKET)
        self.assertEqual(loc.s3_uri, f's3://{TEST_BUCKET}')
        self.assertTrue(loc.is_dir_path())

        # Test with bucket, prefix
        loc = S3Location(TEST_BUCKET, 'test-folder1/test_folder2/', 'us-region-1')
        self.assertEqual(loc.s3_uri, f's3://{TEST_BUCKET}/test-folder1/test_folder2/')
        self.assertEqual(loc.https_url, f'https://s3.us-region-1.amazonaws.com/{TEST_BUCKET}/test-folder1/test_folder2/')
        self.assertTrue(loc.is_dir_path())

        # Test with bucket, key, region
        loc = S3Location(TEST_BUCKET, 'test-folder/test_file.ext', 'us-region-1')
        self.assertEqual(loc.s3_uri, f's3://{TEST_BUCKET}/test-folder/test_file.ext')
        self.assertEqual(loc.https_url, f'https://s3.us-region-1.amazonaws.com/{TEST_BUCKET}/test-folder/test_file.ext')
        self.assertFalse(loc.is_dir_path())
        self.assertTrue(loc.join('').is_dir_path())

        # Test manipulations
        loc = S3Location(TEST_BUCKET, 'test-folder1/', 'us-region-1').join('test_folder2', 'test_file.ext')
        self.assertEqual(
            loc.https_url,
            f'https://s3.us-region-1.amazonaws.com/{TEST_BUCKET}/test-folder1/test_folder2/test_file.ext'
        )
        prefix, name = loc.split_key()
        self.assertEqual(prefix, 'test-folder1/test_folder2')
        self.assertEqual(name, 'test_file.ext')

        self.assertEqual(loc.parent.join(name), loc)
        self.assertEqual(loc.parent / name, loc)

        loc = S3Location(TEST_BUCKET, 'test-folder1/', 'us-region-1').join('test_folder2', '', '')
        self.assertEqual(
            loc.https_url,
            f'https://s3.us-region-1.amazonaws.com/{TEST_BUCKET}/test-folder1/test_folder2/'
        )

        # The following should not be sensitive to the forward slash at the end
        self.assertEqual(S3Location(f's3://{TEST_BUCKET}/').split_key(), ('', ''))
        self.assertEqual(S3Location(f's3://{TEST_BUCKET}/'), f's3://{TEST_BUCKET}')
        self.assertEqual(S3Location(f's3://{TEST_BUCKET}/dir/subdir/').join(''),
                         f's3://{TEST_BUCKET}/dir/subdir')

    @mock_aws
    def test_read_write_ops(self):
        # moto (aws mock) requires the bucket be created before use
        s3._client.create_bucket(Bucket=TEST_BUCKET)

        s3.config(verbose=VERBOSE)
        tests = [
            _create_test_files,
            _test_single_uploads,
            _test_delete,
            _test_recursive_uploads,
            _test_list_objects,
            _test_list_dir,
            _test_download,
            _test_write_read,
        ]
        _vprint('Running tests...\n')
        try:
            for test in tests:
                _vprint(f'----- {test.__name__} -----')
                test()
                _vprint()
        except Exception:
            raise
        finally:
            _cleanup_local()


if __name__ == '__main__':
    arg_parser = get_arg_parser()
    args = arg_parser.parse_args()
    if args.verbose:
        VERBOSE = True
    run_tests(args, TestS3Utils)



from .base import Cache, CacheCore, CachedItemMeta, Serializer, check_meta_mismatches
from .local_cache import LocalDiskCache
from .memory_cache import MemoryCache
from .multi_cache import MultiCache
from .s3_cache import S3Cache
from .setup_helpers import setup_cache



from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, UTC
from typing import Iterable, Iterator, Optional, Self

_NotSpecified = object()


def raise_not_found(key) -> KeyError:
    raise KeyError(f"key '{key}' not found in cache")


@dataclass
class CachedItemMeta:
    key: str
    path: Optional[str]
    size: Optional[int]
    timestamp: Optional[datetime]

    _ts_format = '%Y-%m-%d %H:%M:%S.%f %z'

    def __post_init__(self):
        if isinstance(self.timestamp, float):
            self.timestamp = datetime.fromtimestamp(self.timestamp, UTC)
        if isinstance(self.timestamp, str):
            self.timestamp = datetime.strptime(self.timestamp, self._ts_format)

    def to_dict(self) -> dict:
        return {
            'key': self.key,
            'path': self.path,
            'size': self.size,
            'timestamp': self.timestamp.strftime(self._ts_format) if self.timestamp else None,
        }


class CacheCore(metaclass=ABCMeta):
    def __init__(self, active=True, read_only=False):
        self._read_only = read_only
        self._active = active

    @abstractmethod
    def __contains__(self, key):
        pass

    @abstractmethod
    def get(self, key, default=_NotSpecified):
        pass

    @abstractmethod
    def get_meta(self, key) -> CachedItemMeta:
        pass

    @abstractmethod
    def put(self, key, obj, update=False, meta: Optional[CachedItemMeta] = None):
        pass

    @abstractmethod
    def remove(self, key):
        pass

    @abstractmethod
    def keys(self):
        pass

    def __getitem__(self, key: str):
        return self.get(key)

    def __setitem__(self, key: str, obj) -> None:
        self.put(key, obj)

    def __delitem__(self, key: str) -> None:
        self.remove(key)

    def is_active(self) -> bool:
        return self._active

    def activate(self):
        self._active = True
        return self

    def deactivate(self):
        self._active = False
        return self

    def is_read_only(self) -> bool:
        return self._read_only

    def set_read_only(self, value: bool) -> Self:
        self._read_only = value
        return self

    def update(self, key: str, obj):
        if key not in self:
            raise_not_found(key)
        self.put(key, obj, update=True)

    def pop(self, key: str, default=_NotSpecified):
        obj = self.get(key, default=default)
        self.remove(key)
        return obj

    def list(self) -> dict[str, CachedItemMeta]:
        items_meta = {}
        for key in self.keys():
            items_meta[key] = self.get_meta(key)
        return items_meta

    def subcache(self, *args, **kwargs) -> Self:
        raise NotImplementedError(f"`subcache` method is not defined for class {type(self).__name__}")


class CacheReaderWriter(metaclass=ABCMeta):
    @abstractmethod
    def read(self, key: str):
        pass

    @abstractmethod
    def get_meta(self, key: str) -> CachedItemMeta:
        pass

    @abstractmethod
    def write(self, key: str, obj, meta=None) -> CachedItemMeta:
        pass

    @abstractmethod
    def delete(self, key: str, meta_only=False):
        pass


class CacheCatalog(metaclass=ABCMeta):
    @abstractmethod
    def contains(self, key: str):
        pass

    @abstractmethod
    def add(self, key: str, info: CachedItemMeta):
        pass

    @abstractmethod
    def get(self, key: str) -> CachedItemMeta:
        pass

    @abstractmethod
    def remove(self, key: str):
        pass

    @abstractmethod
    def keys(self):
        pass

    @abstractmethod
    def items(self):
        pass


_CatalogT = dict[str, CachedItemMeta]
_CatalogBuilderT = Callable[[], _CatalogT]


class CacheDictCatalog(CacheCatalog):
    def __init__(
            self,
            catalog: Optional[dict[str, CachedItemMeta]] = None,
            catalog_builder: Optional[_CatalogBuilderT] = None,
    ):
        self._catalog = catalog
        if catalog_builder is None:
            catalog_builder = dict
        self._build_catalog = catalog_builder
        if self._catalog is None:
            self.rebuild()

    def rebuild(self):
        self._catalog = self._build_catalog().copy()

    def contains(self, key: str):
        return str(key) in self._catalog

    def add(self, key: str, meta: CachedItemMeta):
        self._catalog[str(key)] = meta

    def get(self, key: str) -> CachedItemMeta:
        return self._catalog[str(key)]

    def remove(self, key: str):
        self._catalog.pop(str(key))

    def keys(self):
        return list(self._catalog.keys())

    def items(self) -> Iterator[tuple[str, CachedItemMeta]]:
        return iter(self._catalog.items())


class Cache(CacheCore, metaclass=ABCMeta):
    def __init__(self, catalog: CacheCatalog, reader_writer: CacheReaderWriter, active=True, read_only=False):
        super().__init__(active=active, read_only=read_only)
        self._catalog = catalog
        self._reader_writer = reader_writer

    @classmethod
    @abstractmethod
    def create(cls, *args, **kwargs):
        pass

    def __contains__(self, key: str) -> bool:
        if not self.is_active():
            return False
        return self._catalog.contains(key)

    def get(self, key: str, default=_NotSpecified):
        if not self.is_active() or key not in self:
            if default is _NotSpecified:
                raise_not_found(key)
            else:
                return default
        return self._reader_writer.read(key)

    def get_meta(self, key: str) -> CachedItemMeta:
        if not self.is_active() or key not in self:
            raise_not_found(key)
        return self._catalog.get(key)

    def put(self, key: str, obj, *, update=False, meta: Optional[CachedItemMeta] = None) -> Self:
        if self.is_read_only():
            raise TypeError('Cache is in read only mode')
        if self.is_active():
            if key in self and not update:
                raise ValueError(f"key '{key}' already exists in cache; use 'update' to overwrite")
            meta = self._reader_writer.write(key, obj, meta=meta)
            self._catalog.add(key, meta)
        else:
            print(f'WARNING: {type(self).__name__} is not active')
        return self

    def remove(self, key: str, meta_only=False) -> Self:
        if self.is_active() and key in self:
            if self.is_read_only():
                raise TypeError('Cache is in read only mode')
            self._catalog.remove(key)
            self._reader_writer.delete(key, meta_only=meta_only)
        return self

    def remove_meta(self, key: str) -> Self:
        return self.remove(key, meta_only=True)

    def keys(self) -> list[str]:
        if not self.is_active():
            return []
        else:
            return list(self._catalog.keys())

    def _repr_params(self) -> list[str]:
        params = [f'{len(self.keys())} items']
        if not self.is_active():
            params.append('NOT ACTIVE')
        if self.is_read_only():
            params.append('READ ONLY')
        return params

    def __repr__(self):
        return f"{type(self).__name__}({', '.join(self._repr_params())})"


def check_meta_mismatches(caches: Iterable[CacheCore], key=None) -> dict[str, tuple[CachedItemMeta, ...]]:
    if key is not None and not isinstance(key, str):
        raise TypeError('key is not a string')
    for cache in caches:
        if not isinstance(cache, CacheCore):
            raise TypeError('caches must be an iterable containing Cache instances')
    mismatches = {}
    if key is None:
        keys = set(k for cache in caches for k in cache.keys())
    else:
        keys = [key]
    for k in keys:
        metas = [cache.get_meta(k) for cache in caches if k in cache]
        if len(metas) > 1:
            first_meta = metas[0]
            if any(meta != first_meta for meta in metas[1:]):
                mismatches[k] = tuple(metas)
    return mismatches


class Serializer(metaclass=ABCMeta):
    default_ext = None

    def __init__(self, ext=None):
        if ext is None:
            self.ext = self.default_ext
        else:
            self.ext = ext

    @abstractmethod
    def read(self, path):
        pass

    @abstractmethod
    def write(self, path, obj):
        pass



import json
import os
import pickle
from datetime import datetime, UTC
from functools import partial
from glob import glob
from pathlib import Path
from typing import Optional, Self

from yes3.caching.base import Cache, CacheDictCatalog, CachedItemMeta, Serializer, CacheReaderWriter


class PickleSerializer(Serializer):
    default_ext = 'pkl'

    def read(self, path):
        with open(path, 'rb') as f:
            return pickle.load(f)

    def write(self, path, obj):
        os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)
        with open(path, 'wb') as f:
            pickle.dump(obj, f)


class JsonSerializer(Serializer):
    default_ext = 'json'

    def read(self, path) -> dict:
        with open(path, 'r') as f:
            return json.load(f)

    def write(self, path, obj: dict):
        os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)
        with open(path, 'w') as f:
            json.dump(obj, f)


class JsonMetaSerializer(JsonSerializer):
    default_ext = 'meta'

    def read(self, path) -> CachedItemMeta:
        meta_dict = super().read(path)
        return CachedItemMeta(**meta_dict)

    def write(self, path, meta: CachedItemMeta):
        super().write(path, meta.to_dict())


def _get_serializer(serializer: str | Serializer, ext=None) -> Serializer:
    if isinstance(serializer, type):
        serializer = serializer(ext)

    if isinstance(serializer, str):
        if serializer.lstrip('.').lower() in {'pkl', 'pickle'}:
            return PickleSerializer(ext)
        elif serializer.lstrip('.').lower() == 'json':
            return JsonSerializer(ext)
        else:
            raise NotImplementedError(f"Serializer not implemented for file type '{serializer}'")
    elif isinstance(serializer, Serializer):
        if ext is not None:
            serializer.ext = ext
        return serializer
    else:
        raise TypeError(
            f'file_type must be a string or a Serializer subclass, but got type {type(serializer).__name__}')


def _with_ext(path, ext: Optional[str]):
    if ext is None:
        return path
    if not ext.startswith('.'):
        ext = f'.{ext}'
    path_str = str(path)
    if path_str.endswith(ext):
        return path
    else:
        try:
            return type(path)(path_str + ext)
        except (ValueError, TypeError):
            return Path(path_str + ext)


class LocalReaderWriter(CacheReaderWriter):
    def __init__(
            self, path: str | Path,
            object_serializer: str | Serializer = PickleSerializer(),
            meta_serializer: str | Serializer = JsonMetaSerializer(),
    ):
        self.path = Path(path)
        self.obj_serializer = _get_serializer(object_serializer)
        self.meta_serializer = _get_serializer(meta_serializer)

    def clone(self, path: str | Path) -> Self:
        return type(self)(path, object_serializer=self.obj_serializer, meta_serializer=self.meta_serializer)

    def key2path(self, key: str, meta=False) -> Path:
        if meta:
            return self.path / _with_ext(key, self.meta_serializer.ext)
        else:
            return self.path / _with_ext(key, self.obj_serializer.ext)

    def path2key(self, path: str | Path) -> str:
        path = Path(path)
        rel_path = path.relative_to(self.path)
        return rel_path.stem

    def read(self, key: str):
        path = self.key2path(key)
        print(f"Reading cached item '{key}' at {path}")
        return self.obj_serializer.read(path)

    def _build_meta(self, path, key=None) -> CachedItemMeta:
        if key is None:
            key = self.path2key(path)
        file_stat = os.stat(path)
        rel_path = path.relative_to(self.path)
        return CachedItemMeta(
            key=key,
            path=str(rel_path),
            size=file_stat.st_size,
            timestamp=datetime.fromtimestamp(file_stat.st_mtime, UTC),
        )

    def get_meta(self, key: str, rebuild=False) -> CachedItemMeta:
        if rebuild:
            obj_path = self.key2path(key)
            meta_path = self.key2path(key, meta=True)
            meta = self._build_meta(path=obj_path, key=key)
            self.meta_serializer.write(meta_path, meta)
        else:
            meta_path = self.key2path(key, meta=True)
            meta = self.meta_serializer.read(meta_path)
        return meta

    def write(self, key: str, obj, meta: Optional[CachedItemMeta] = None) -> CachedItemMeta:
        obj_path = self.key2path(key)
        print(f"Caching item '{key}' at {obj_path}")
        self.obj_serializer.write(obj_path, obj)

        meta_path = self.key2path(key, meta=True)
        if meta is None:
            meta = self._build_meta(path=obj_path, key=key)
        self.meta_serializer.write(meta_path, meta)
        return meta

    def delete(self, key: str, meta_only=False):
        path = self.key2path(key)
        meta_path = self.key2path(key, meta=True)
        if meta_only:
            print(f"Deleting cached item '{key}' metadata at {meta_path}")
        else:
            print(f"Deleting cached item '{key}' at {path}")
            os.remove(path)
        os.remove(meta_path)


class LocalDiskCache(Cache):
    @staticmethod
    def _build_catalog_dict(reader_writer: LocalReaderWriter, rebuild_missing_meta=False) -> dict:
        catalog_dict = {}
        if os.path.exists(reader_writer.path):
            data_ext = reader_writer.obj_serializer.ext.lstrip('.')
            meta_ext = reader_writer.meta_serializer.ext.lstrip('.')
            data_files = glob(str(reader_writer.path / f'*.{data_ext}'))
            meta_files = glob(str(reader_writer.path / f'*.{meta_ext}'))
            data_map = {Path(p).stem: p for p in data_files}
            meta_map = {Path(p).stem: p for p in meta_files}
            if data_map.keys() != meta_map.keys():
                if rebuild_missing_meta:
                    print(f'WARNING: data and metadata files are not aligned for cache at {reader_writer.path}, '
                          'rebuilding missing metadata files')
                else:
                    raise RuntimeError(f'data and metadata files are not aligned for cache at {reader_writer.path}')
            for key in data_map.keys():
                catalog_dict[key] = reader_writer.get_meta(key, rebuild=(key not in meta_map and rebuild_missing_meta))
        if len(catalog_dict.keys()) > 0:
            print(f'{len(catalog_dict.keys())} cached items discovered at {reader_writer.path}')
        return catalog_dict

    @classmethod
    def create(
            cls,
            path: str | Path,
            obj_serializer: str | Serializer = PickleSerializer(),
            meta_serializer: str | Serializer = JsonMetaSerializer(),
            reader_writer: Optional[CacheReaderWriter] = None,
            rebuild_missing_meta=False,
            **kwargs,
    ):
        if reader_writer is None:
            reader_writer = LocalReaderWriter(path, obj_serializer, meta_serializer)
        elif not isinstance(reader_writer, LocalReaderWriter):
            raise TypeError(f'`reader_writer` must be a {LocalReaderWriter.__name__} instance')
        elif reader_writer.path != path:
            reader_writer = reader_writer.clone(path)
        catalog_builder = partial(cls._build_catalog_dict, reader_writer=reader_writer,
                                  rebuild_missing_meta=rebuild_missing_meta)
        catalog = CacheDictCatalog(catalog_builder=catalog_builder)
        return cls(catalog, reader_writer, **kwargs)

    @property
    def path(self) -> Path:
        return self._reader_writer.path

    def subcache(self, rel_path: str) -> Self:
        path = self.path / rel_path
        kwargs = dict(active=self.is_active(), read_only=self.is_read_only())
        return type(self).create(path, reader_writer=self._reader_writer, **kwargs)

    def clear(self, force=False) -> Self:
        if self.is_active() and len(self.keys()) > 0:
            if not force:
                raise RuntimeError(f'Clearing this {type(self).__name__} ({self.path}) requires specifying force=True')
            print(f'Deleting {len(self.keys())} item(s) from cache at {self.path}')
            for key in self.keys():
                self.remove(key)
            new_cache = type(self).create(self.path, reader_writer=self._reader_writer)
            self.__init__(new_cache._catalog, new_cache._reader_writer, active=self._active, read_only=self._read_only)
        return self

    def clear_meta(self, force=False) -> Self:
        if self.is_active() and len(self.keys()) > 0:
            if not force:
                raise RuntimeError(f'Clearing this {type(self).__name__} metadata ({self.path}) requires specifying '
                                   'force=True')
            print(f'Deleting {len(self.keys())} item(s) from cache at {self.path}')
            for key in self.keys():
                self.remove(key, meta_only=True)
        return self

    def _repr_params(self) -> list[str]:
        params = super()._repr_params()
        params.insert(0, str(self.path))
        return params
